#!/bin/env python

#######################################################################
#  Copyright (C) 2023 Vinh Tran
#
#  This file is part of FAS. It is used to calculate FAS scores for
#  input that contains pairwise proteins from different pair of taxa
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


import os
import sys
import argparse
import multiprocessing as mp
from sys import argv
from greedyFAS.annoFAS import annoFAS, annoModules
from greedyFAS.mainFAS import fasInput, fasOutput, greedyFAS
from greedyFAS.mainFAS.fasInput import read_json
from greedyFAS import calcFAS
from greedyFAS import complexityFAS
from pkg_resources import get_distribution
import multiprocessing as mp
import subprocess
from tqdm import tqdm
import json
import shutil
import glob
from datetime import date
import random


def get_options():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    general = parser.add_argument_group('general arguments')
    inargs = parser.add_argument_group('input arguments')
    outargs = parser.add_argument_group('output arguments')
    annotation = parser.add_argument_group('annotation arguments')
    weighting = parser.add_argument_group('weighting arguments')
    thresholds = parser.add_argument_group('threshold arguments')
    obscure = parser.add_argument_group('obscure arguments')
    other = parser.add_argument_group('other arguments')
    parser.add_argument('--version', action='version', version=str(version))
    required.add_argument("--input", default=None, type=str, required=True,
                          help="Input file (see https://github.com/BIONF/FAS/wiki/File-Formats)")
    required.add_argument("-a", "--annotation_dir", default=None, type=str, required=True,
                          help='working directory, all annotations are be stored here')
    required.add_argument("-o", "--out_dir", default=None, type=str, required=True,
                          help="output directory, all outputfiles will be stored here")
    general.add_argument("--bidirectional", action="store_true",
                         help="calculate both scoring directions")
    general.add_argument('--cpus', help='number of cores', type=int, action='store', default=0)
    annotation.add_argument("-f", "--eFeature", default="0.001", type=float,
                            help="eValue cutoff for PFAM/SMART domain, applied during annotation but also during "
                                 "calculation")
    annotation.add_argument("-i", "--eInstance", default="0.01", type=float,
                            help="eValue cutoff for PFAM/SMART instances, applied during annotation but also during "
                                 "calculation")
    annotation.add_argument('--eFlps', help='eValue cutoff for fLPS', action='store', default=0.0000001, type=float)
    annotation.add_argument('--org', help='Organism of input sequence(s) for SignalP search',
                            choices=['euk', 'gram+', 'gram-'], action='store', default='euk', type=str)
    annotation.add_argument("--toolPath", dest="toolPath", default='', type=str,
                            help="Path to Annotion tool directory created with fas.setup")
    weighting.add_argument("-g", "--weight_correction", default="loge", type=str,
                           help="Function applied to the frequency of feature types during weighting, options are "
                                "linear(no function), loge(natural logarithm[Default]), log10(base-10 logarithm), "
                                "root4(4th root) and root8(8th root).")
    weighting.add_argument("-x", "--weight_constraints", default=None, type=str,
                           help="Apply weight constraints via constraints file, by default there are no constraints. "
                                "Please look at the FAS wiki pages for templates for the constraints file")
    inargs.add_argument("--oldJson", default='', type=str,
                        help="Input old FAS output in JSON format. Only new pairs will be considered!")
    inargs.add_argument("-d", "--featuretypes", default='', type=str,
                        help="inputfile that contains the tools/databases used to predict features. Please look at the "
                             "FAS wiki pages for templates of the the featuretypes input file")
    inargs.add_argument("--extra_annotation", default=None, nargs='*', type=str,
                        help="give naming conventions for extra annotation files, eg. if the file name for the seed is "
                             "someseed.json than the extra annotation files should be named someseed_"
                             "[EXTRA_ANNOTATION].json, these extra files need to exist for seed, query and the "
                             "references (if given)")
    inargs.add_argument("--pairLimit", default=0, type=int,
                        help="Limit the number of pairs to be considered")
    outargs.add_argument("--raw", dest="raw", action="store_true",
                         help="print FAS scores to terminal")
    outargs.add_argument("--tsv", dest="tsv", action="store_true",
                         help="deactivates creation of the tsv output, either use together with --raw or "
                              "--phyloprofile")
    outargs.add_argument("--json", dest="json", action="store_true",
                         help="create json output")
    outargs.add_argument("--mergeJson", action="store_true",
                        help="merge multiple json outputs")
    outargs.add_argument("--outName", default=None, type=str,
                        help="name for merged output file, if none is given the name will be created by the date")
    outargs.add_argument("--phyloprofile", dest="phyloprofile", default=None, type=str,
                         help="activate phyloprofile output, needs mapping file for all query proteins")
    outargs.add_argument("--domain", dest="domain", action="store_false",
                         help="deactivate .domains output")
    outargs.add_argument("--no_config", dest="config", action="store_false",
                         help="deactivate *_config.yml output")
    outargs.add_argument('--force', help='Force override output files', action='store_true')
    outargs.add_argument('--silentOff', help='Turn on STDOUT (not recommended for parallel run)', action='store_true')
    outargs.add_argument('--keep', help='Keep split input files', action='store_true')
    thresholds.add_argument("-c", "--max_overlap", dest="max_overlap", default=0, type=int,
                            help="maximum size overlap allowed, default is 0 amino acids")
    thresholds.add_argument("--max_overlap_percentage", dest="max_overlap_percentage", default=0.4, type=float,
                            help="defines how much percent of a feature the overlap is allowed to cover, default "
                                 "is 0.4 (40%%)")
    thresholds.add_argument("-t", "--priority_threshold", type=int, default=30,
                            help="Change to define the feature number threshold for activating priority mode in the "
                                 "path evaluation. default=30")
    thresholds.add_argument("-m", "--max_cardinality", default=500, type=int,
                            help="Change to define the threshold for the maximal cardinality (number) of feature paths "
                                 "in a graph. If max. cardinality is exceeded the priority mode will be used to for "
                                 "the path evaluation. default=500")
    thresholds.add_argument("--paths_limit", default=0, type=int,
                            help="Specify number of maximum paths to be considered (10^n). If this threshold is exceeded, "
                                 "the corresponding protein will be ignored. Default: 0 for no limit")
    obscure.add_argument("--priority_mode", action='store_false',
                         help="deactivates the greedy strategy priority mode for larger architectures, NOT RECOMMENDED")
    obscure.add_argument("--timelimit", default=3600, type=int,
                         help="Sets a soft time limit in seconds for the calculation between a pair of proteins,"
                              "This limit does not necessarily represent the actual runtime. It only stops the "
                              "exhaustive strategy through the architecture and activate priority mode. "
                              "This option is only relevant if the option [--priority_mode] is set")
    obscure.add_argument("-w", "--score_weights", nargs=3, default=[0.7, 0.0, 0.3], type=float,
                         help="Defines how the three scores MS, CS and PS are weighted, takes three float arguments, "
                              "sum should be 1.0, the default is 0.7, 0.0, 0.3")
    obscure.add_argument("--empty_as_1", action='store_true',
                         help="If both proteins have no features, score 1.0 instead of the default 0.0")
    other.add_argument("--check_anno", action='store_true',
                         help="Checking for existing annotations")
    args = parser.parse_args()
    return args


def check_anno(in_file, annotation_dir):
    """ Check if annotations for all input proteins are present
    """
    # get list of taxa and their protein IDs
    input_dict = {}
    with open(in_file, 'r') as fr:
        for line in fr:
            tmp = line.strip().split()
            if len(tmp) == 4:
                if not tmp[1] in input_dict:
                    input_dict[tmp[1]] = []
                input_dict[tmp[1]].append(tmp[0])
                if not tmp[3] in input_dict:
                    input_dict[tmp[3]] = []
                input_dict[tmp[3]].append(tmp[2])
    # check missing anno
    missing_dict = {}
    for taxon in input_dict:
        anno_dict = read_json(f'{annotation_dir}/{taxon}.json')
        for id in input_dict[taxon]:
            if not id in anno_dict['feature']:
                if not taxon in missing_dict:
                    missing_dict[taxon] = []
                missing_dict[taxon].append(id)
    return(missing_dict)


def calc_path_number(prot_id, anno_file):
    """ Calculate number of feature paths for a given protein ID in a anno file
    """
    anno_dict = {}
    clan_dict = {}
    anno_dict = read_json(anno_file)
    proteome = anno_dict["feature"]
    clan_dict.update(anno_dict["clan"])
    option_dict = {}
    pathconfigfile = os.path.realpath(__file__).replace('calcFASmulti.py', 'pathconfig.txt')
    with open(pathconfigfile) as f:
        toolpath = f.readline().strip()
    option_dict['input_linearized'], option_dict['input_normal'] = fasInput.featuretypes(toolpath + '/'
                                                                                             + 'annoTools.txt')
    option_dict["max_overlap"] = 0
    option_dict["max_overlap_percentage"] = 0.4
    option_dict["eFeature"] = 0.001
    option_dict["eInstance"] = 0.01

    (path_number, greedy_comp, graph) = complexityFAS.calc_complex(prot_id, proteome, option_dict, clan_dict)
    return(path_number)


def filter_oldJson(old_json, in_file):
    """ Function to check if a FAS scores for input protein pair already exists
    Return a list of lines from in_file containing new pairs
    """
    old_results = read_json(old_json)
    new_in_file = []
    with open(in_file, 'r') as fr:
        for line in fr:
            tmp = line.strip().split()
            if not f'{tmp[0]}_{tmp[2]}' in old_results and not f'{tmp[2]}_{tmp[0]}' in old_results:
                new_in_file.append(line.strip())
    return(new_in_file)


def get_prot_for_taxpair(opts):
    """ Create file contaning protein pairs for each pair of taxa
    If a path limit is given, only protein that have less number of paths will
    by saved
    """
    (line, paths_limit, annotation_dir, out_dir, out_name) = opts # line = id1  tax1  id2  tax2
    tmp = line.strip().split()
    taxa_pair = ''
    if len(tmp) == 4:
        taxa_pair = f'{tmp[1]}#{tmp[3]}'
        # if paths_limit > 0:
        #     path_p1 = calc_path_number(tmp[0], f'{annotation_dir}/{tmp[1]}.json')
        #     path_p2 = calc_path_number(tmp[2], f'{annotation_dir}/{tmp[3]}.json')
        #     if path_p1 > 10**paths_limit or path_p2 > 10**paths_limit:
        #         return('')
        fp = open(f'{out_dir}/{out_name}_split_inputs/{taxa_pair}.txt', 'a+')
        fp.write(f'{tmp[0]}\t{tmp[2]}\n')
        fp.close()
    return(taxa_pair)


def create_jobs(in_file, args, annotation_dir, out_dir, out_name, toolpath, cpus):
    """ Create jobs for running calcFas
    """
    tax_pairs = []
    if os.path.exists(f'{out_dir}/{out_name}_split_inputs'):
        if args.force:
            shutil.rmtree(f'{args.out_dir}/{out_name}_split_inputs')
        else:
            sys.exists(f'{out_dir}/{out_name}_split_inputs exists! Please use --force to overwrite or manually delete that folder to continue')
    os.makedirs(f'{out_dir}/{out_name}_split_inputs', exist_ok=True)
    with open(in_file, 'r') as fr:
        lines = fr.read().splitlines()
        if args.oldJson:
            if os.path.exists(os.path.abspath(args.oldJson)):
                lines = filter_oldJson(args.oldJson, in_file)
            else:
                print(f'WARNING: {args.oldJson} not found!')
        if len(lines) > 0:
            parse_taxa_jobs = []
            if args.pairLimit > 0:
                if args.silentOff:
                    print(f"NOTE: only random {args.pairLimit} pairs will be calculated!")
                if args.pairLimit < len(lines):
                    lines = random.sample(lines, args.pairLimit)
            for line in lines:
                parse_taxa_jobs.append((line, args.paths_limit, annotation_dir, out_dir, out_name))
            if cpus == 1:
                for l in parse_taxa_jobs:
                    tax_pairs.append(get_prot_for_taxpair(l))
            else:
                pool = mp.Pool(cpus)
                for _ in tqdm(pool.imap_unordered(get_prot_for_taxpair, parse_taxa_jobs), total=len(parse_taxa_jobs)):
                    tax_pairs.append(_)
                pool.close()

        tax_pairs = list(set(tax_pairs))
        tax_pairs = list(filter(None, tax_pairs))
        jobs = []
        for tax_pair in tax_pairs:
            args_dict_new = {}
            args_dict = vars(args)
            for item in args_dict:
                args_dict_new[item] = args_dict[item]
            args_m = argparse.Namespace(**args_dict_new)
            tmp = tax_pair.split('#')
            args_m.seed = f'{tmp[0]}.json'
            args_m.query = f'{tmp[1]}.json'
            args_m.pairwise = f'{out_dir}/{out_name}_split_inputs/{tax_pair}.txt'
            args_m.out_name = tax_pair
            args_m.query_id = None
            args_m.seed_id = None
            args_m.ref_proteome = None
            args_m.ref_2 = None
            if args.silentOff:
                args_m.silent = False
            else:
                args_m.silent = True
            # check existing output files
            out_file = f'{args_m.out_dir}/{args_m.out_name}.tsv'
            if args.json:
                out_file = f'{args_m.out_dir}/{args_m.out_name}.json'
            if os.path.exists(out_file):
                if not args.force:
                    print(f'Output file {os.path.abspath(out_file)} exists! Use --force if you want to overwrite it!')
                    continue
            jobs.append([args_m, toolpath])
    return(jobs)


def check_json_output(out_name, out_dir, old_json, force):
    """ Check if outName.json file exists
    Or if any other .json file present in out_dir
    """
    # check out_name.json exists
    if os.path.exists(f'{out_dir}/{out_name}.json'):
        if force:
            os.remove(f'{out_dir}/{out_name}.json')
        else:
            sys.exit(f'{out_dir}/{out_name}.json exists! Use --force to overwrite or move is to other directory.')
    # check other json files. If present, move them to out_dir/old/
    json_file = glob.glob(os.path.join(out_dir, '*.json'))
    if len(json_file) > 0:
        os.makedirs(f'{out_dir}/{out_name}_old', exist_ok = True)
        for jf in json_file:
            if not os.path.abspath(jf) == os.path.abspath(old_json):
                shutil.move(jf, f"{out_dir}/{out_name}_old/{jf.split('/')[-1]}")


def main():
    args = get_options()
    toolpath = args.toolPath
    if toolpath == '':
        pathconfigfile = os.path.realpath(__file__).replace('calcFASmulti.py', 'pathconfig.txt')
        with open(pathconfigfile) as f:
            toolpath = f.readline().strip()
    else:
        toolpath = os.path.abspath(args.toolPath)
    if not os.path.isdir(args.annotation_dir):
        sys.exit(f'ERROR: {args.annotation_dir} not found!')
    if not os.path.isdir(args.out_dir):
        os.mkdir(args.out_dir)
    if args.force:
        print('WARNING: --force option is used. Old output files will be overwritten!')

    if args.check_anno:
        print('==> checking annotations...')
        missing_dict = check_anno(args.input, args.annotation_dir)
        if len(missing_dict) > 0:
            sys.exit(f'ERROR: Annotations for the following proteins are missing!\n{missing_dict}')

    out_name = args.outName
    out_dir = os.path.abspath(args.out_dir)
    if not args.outName:
        today = date.today()
        out_name = f"fas_{today.strftime('%y%m%d')}"

    if args.mergeJson:
        check_json_output(out_name, out_dir, args.oldJson, args.force)


    print('==> preparing jobs...')
    if args.cpus == 0:
        cpus = mp.cpu_count()-1
    else:
        if args.cpus > mp.cpu_count():
            cpus = mp.cpu_count()-1
        else:
            cpus = args.cpus
    jobs = create_jobs(args.input, args, args.annotation_dir, args.out_dir, out_name, toolpath, cpus)


    if len(jobs) > 0:
        print('==> calculating FAS scores...')
        if cpus == 1:
            for j in jobs:
                calcFAS.fas(j)
        else:
            pool = mp.Pool(cpus)
            for _ in tqdm(pool.imap_unordered(calcFAS.fas, jobs), total=len(jobs)):
                pass
            pool.close()
    else:
        sys.exit(f'==> DONE! All pairwise FAS scores can be found in {args.oldJson}!')


    if args.mergeJson:
        out_dir = os.path.abspath(args.out_dir)
        print(f'==> merge outputs into {out_name}...')
        os.makedirs(f'{out_dir}/{out_name}_tmp', exist_ok = True)
        for json_file in glob.glob(os.path.join(out_dir, '*.json')):
            shutil.move(json_file, f"{out_dir}/{out_name}_tmp/{json_file.split('/')[-1]}")
            if args.oldJson:
                if os.path.abspath(json_file) == os.path.abspath(args.oldJson):
                    shutil.move(f"{out_dir}/{out_name}_tmp/{json_file.split('/')[-1]}", args.oldJson)

        cmd = f'fas.mergeJson -i {out_dir}/{out_name}_tmp/ -n {out_name} -o {out_dir}'
        try:
            merged_out = subprocess.run([cmd], shell=True, capture_output=True, check=True)
        except:
            sys.exit(f'Error running\n{cmd}')
        # if args.oldJson:
        #     if os.path.exists(f"{out_dir}/{out_name}_old/{args.oldJson.split('/')[-1]}"):
        #         shutil.move(f"{out_dir}/{out_name}_old/{args.oldJson.split('/')[-1]}", os.path.abspath(args.oldJson))
        #     if os.path.exists(os.path.abspath(args.oldJson)):
        #         old_json_path = os.path.abspath(args.oldJson)
        #         update_cmd = f'fas.updateJson --old {os.path.abspath(args.oldJson)} --new {out_dir}/{out_name}.json'
        #         try:
        #             update_out = subprocess.run([update_cmd], shell=True, capture_output=True, check=True)
        #         except:
        #             sys.exit(f'Error running\n{update_cmd}')
        #     else:
        #         print(f'WARNING: {args.oldJson} not found!')
        if os.path.exists(f'{out_dir}/{out_name}_old'):
            for jf in glob.glob(os.path.join(f'{out_dir}/{out_name}_old', '*.json')):
                shutil.move(jf, f"{out_dir}/{jf.split('/')[-1]}")
            shutil.rmtree(f'{args.out_dir}/{out_name}_old')


    if not args.keep:
        shutil.rmtree(f'{args.out_dir}/{out_name}_split_inputs')
        if args.mergeJson:
            shutil.rmtree(f'{args.out_dir}/{out_name}_tmp')
    print('==> DONE!')

if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2019 Julian Dosch
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


from operator import itemgetter
import os
import multiprocessing
from functools import partial
from copy import deepcopy
from tqdm import tqdm
import itertools
import sys
from time import sleep
from greedyFAS.mainFAS.fasInput import read_json
from greedyFAS.mainFAS.fasInput import check_version
from greedyFAS.mainFAS.fasOutput import write_domain_out
from greedyFAS.mainFAS.fasOutput import write_tsv_out, write_json_out
from greedyFAS.mainFAS.fasOutput import phyloprofile_out
from greedyFAS.mainFAS.fasScoring import sf_calc_score
from greedyFAS.mainFAS.fasScoring import sf_entire_calc_score
from greedyFAS.mainFAS.fasWeighting import w_weight_const_rescale
from greedyFAS.mainFAS.fasWeighting import w_weight_correction
from greedyFAS.mainFAS.fasWeighting import w_weighting
from greedyFAS.mainFAS.fasWeighting import w_weighting_constraints
from greedyFAS.mainFAS.fasWeighting import w_count_add_domains
from greedyFAS.mainFAS.fasPathing import pb_region_mapper
from greedyFAS.mainFAS.fasPathing import pb_region_paths
from greedyFAS.annoFAS.annoModules import mergeNestedDic


# important vars #             ###  var looks ###
# naming
# seed_proteome = {}           #{("protein_id", {("domain_name", [("START", "STOP")])})}
# query_proteome = {}          #{("protein_id", {("domain_name", [("START", "STOP")])})}
# clan_dict = {}               #{("domain_name", "clan")}
# search_features = {}         #{("F_0", ("domain_name", "POSITION", "Start", "Stop" ))}
# a_s_f = {}                   # additional seed features [non linearized]
# query_features = {}          #{("F_0", ("domain_name", "POSITION", "Start", "Stop" ))}
# a_q_f = {}                   # additional query features [non linearized]
# query_protein = {}           #{("domain_name", ["POSITION_1", "POSITION_2"])}
# query_clans = {}             #{("clan", "INSTANCES")}
# weights = {}                 #{("domain_name", "weight")}
# domain_count = {}            #{("domain", "COUNT")}


# flow control functions


def fc_start(option):
    """Overhead function,
    this function manages the individual functions that read the input files and prepares the data for the main script.

    :param option: dictionary that contains the main option variables of FAS
    """
    clan_dict = {}
    domain_count = {}
    phmm = {}
    option["reverse"] = False
    interprokeys = {}
    v_warning = False
    version = None
    # MS_uni set to 0 when no weighting is conducted
    if option["MS_uni"] == 0:
        domain_count = {}
        for path in option["ref_proteome"]:
            proteome = read_json(path)
            domain_count.update(proteome["count"])
            version, v_warning = check_version(version, proteome, v_warning)
    proteome_list = []
    for path in option["p_path"]:
        proteome = read_json(path)
        proteome_list.append(proteome["feature"])
        clan_dict.update(proteome["clan"])
        if 'interproID' in proteome:
            interprokeys.update(proteome['interproID'])
        if 'length' in proteome:
            phmm.update(proteome['length'])
        version, v_warning = check_version(version, proteome, v_warning)
    seed_proteome = mergeNestedDic(proteome_list)
    proteome_list = []
    for path in option["s_path"]:
        proteome = read_json(path)
        proteome_list.append(proteome["feature"])
        clan_dict.update(proteome["clan"])
        if 'interproID' in proteome:
            interprokeys.update(proteome['interproID'])
        if 'length' in proteome:
            phmm.update(proteome['length'])
        version, v_warning = check_version(version, proteome, v_warning)
    query_proteome = mergeNestedDic(proteome_list)
    if v_warning:
        print('##########\nWARNING: There are version discrepancies in the annotation files. This means that there '
              'might be differences in tool/database versions of the annotation. You should check the annotation files '
              'with "fas.getAnnoVersion"!\n##########')
    if option["weight_correction"]:
        domain_count = w_weight_correction(option["weight_correction"], domain_count)
    for tool in option["input_linearized"]:
        if tool not in seed_proteome[list(seed_proteome)[0]]:
            raise Exception(tool + " is missing in the seed annotation")
        if tool not in query_proteome[list(query_proteome)[0]]:
            raise Exception(tool + " is missing in the query annotation")
    if option["seed_id"]:
        for protid in option["seed_id"]:
            if protid not in seed_proteome:
                raise Exception(protid + " is not in the seed annotation")
    if option["query_id"]:
        for protid in option["query_id"]:
            if protid not in query_proteome:
                raise Exception(protid + " is not in the query annotation")
    if option["bidirectional"]:
        if not option["silent"]:
            print("calculating forward scores...")
        f_results = fc_main(domain_count, seed_proteome, query_proteome, clan_dict, option, interprokeys, phmm)
        r_results = ()
        if not f_results[-1] == 'NA':
            if option["MS_uni"] == 0 and option["ref_2"]:
                domain_count_2 = {}
                for path in option["ref_2"]:
                    domain_count_2.update(read_json(path)["count"])
                if option["weight_correction"]:
                    domain_count_2 = w_weight_correction(option["weight_correction"], domain_count_2)
                option['ref_proteome'] = option['ref_2']
            else:
                domain_count_2 = domain_count
            id_tmp = option["seed_id"]
            option["reverse"] = True
            option["seed_id"] = option["query_id"]
            option["query_id"] = id_tmp
            if option["pairwise"]:
                pairtmp = []
                for pair in option["pairwise"]:
                    pairtmp.append([pair[1], pair[0]])
                option["pairwise"] = pairtmp
            if not option["silent"]:
                print("calculating backward scores...")
            r_results = fc_main(domain_count_2, query_proteome, seed_proteome, clan_dict, option, interprokeys, phmm)

        if option["phyloprofile"]:
            phyloprofile_out(option["outpath"], True, option["phyloprofile"], (f_results, r_results))
        if not option['tsv']:
            write_tsv_out(option["outpath"], True, (f_results, r_results))
        if option['json']:
            write_json_out(option["outpath"], True, (f_results, r_results))
    else:
        if not option["silent"]:
            print("calculating forward scores...")
        results = fc_main(domain_count, seed_proteome, query_proteome, clan_dict, option, interprokeys, phmm)
        if not option["tsv"]:
            write_tsv_out(option["outpath"], False, (results, None))
        if option["json"]:
            write_json_out(option["outpath"], False, (results, None))
        if option["phyloprofile"]:
            phyloprofile_out(option["outpath"], False, option["phyloprofile"], [results, None])


def filter_oldJson(old_json, in_pairs):
    """ Function to check if a FAS scores for input protein pair already exists
    Return a list of new pairs only
    """
    old_results = read_json(old_json)
    new_pairs = []
    for pair in in_pairs:
        if not f'{pair[0]}_{pair[1]}' in old_results and not f'{pair[1]}_{pair[0]}' in old_results:
            new_pairs.append(pair)
    return(new_pairs)


def fc_main(domain_count, seed_proteome, query_proteome, clan_dict, option, interprokeys, phmm):
    """Main function,
    manages linearization and scoring, creates the output
    Function calls: w_weighting_constraints(), w_weighting(), su_lin_query_protein(), pb_graphtraversal(),
                    pb_entire_graphtraversal_priority(),
                    sf_entire_calc_score(), pb_entire_main_nongreedy(), w_weight_const_rescale()

    :param domain_count: dictionary that contains the (adjusted) counts of each feature in the reference proteome
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param query_proteome: dictionary that contains the feature architecture of all query proteins
    :param clan_dict: dictionary that maps features to clans
    :param option: dictionary that contains the main option variables of FAS
    """

    progress = None
    domain_out = None
    results = []
    if option["domain"]:
        if option["reverse"]:
            domain_out = open(option["outpath"] + "_reverse.domains", "w")
        else:
            domain_out = open(option["outpath"] + "_forward.domains", "w")
        domain_out.write('# pairID\torthoID\tseqLen\tfeature\tfStart\tfEnd\tfWeight\tfPath\tinterProID\te-value'
                         + '\tbitScore\tpStart\tpEnd\tpLen\n')

    final_pairs = []
    if option['pairwise']:
        final_pairs = option['pairwise']
    else:
        if option["query_id"]:
            querylist = option["query_id"]
            for query in querylist:
                if query not in query_proteome:
                    raise Exception(query + ' is missing in annotation!')
        else:
            querylist = list(query_proteome.keys())
        if option["seed_id"]:
            seedlist = option["seed_id"]
            for seed in seedlist:
                if seed not in seed_proteome:
                    raise Exception(seed + ' is missing in annotation!')
        else:
            seedlist = list(seed_proteome.keys())
        # create pairwise list
        final_pairs = list(itertools.product(seedlist,querylist))
        if option['old_json']:
            final_pairs = filter_oldJson(option['old_json'], final_pairs)
    if len(final_pairs) > 0:
        if option['progress']:
             progress = tqdm(total=len(final_pairs), file=sys.stdout)

        for pair in final_pairs:
            query = pair[1]
            protein = pair[0]
            ### CHECK QUERY & PROTEIN IN JSON ################
            if query not in query_proteome:
                raise Exception(query + ' is missing in annotation!')
            if protein not in seed_proteome:
                raise Exception(protein + ' is missing in annotation!')
            tmp_query = fc_prep_query(query, domain_count, query_proteome, option, clan_dict)
            if not tmp_query == None:
                tmp_protein = fc_prep_query(protein, 'NA', seed_proteome, option, clan_dict)
                if not tmp_protein == None:
                    query_graph, all_query_paths, lin_query_set, query_features, a_q_f, query_clans, clan_dict = tmp_query[0:7]
                    go_priority, domain_count = tmp_query[7:9]

                    #### WRITE RESULTS ####################
                    results.append(fc_main_sub(protein, domain_count, seed_proteome, option, all_query_paths, query_features,
                                               go_priority, a_q_f, clan_dict, query_graph, query_proteome, query, query_clans,
                                               domain_out, interprokeys, phmm))
                    if option["progress"]:
                        progress.update(1)
                else:
                    results.append((protein, query, ('NA', 'NA', 'NA', 'NA', 'NA'), 'NA'))
            else:
                results.append((protein, query, ('NA', 'NA', 'NA', 'NA', 'NA'), 'NA'))

        if option["progress"]:
            progress.refresh()
            progress.close()
            sleep(1.0)
    else:
        print('==> No new protein pairs need to be calculated!')
    return results


def fc_prep_query(query, domain_count, query_proteome, option, clan_dict):
    go_priority = False
    if option["MS_uni"] == 0:
        domain_count = w_count_add_domains(query, domain_count, query_proteome)
    lin_query_set, query_features, a_q_f, query_clans, clan_dict = su_lin_query_protein(
        query, query_proteome, clan_dict, option)
    tmp_query_graph, path_number = pb_region_paths(pb_region_mapper(
        lin_query_set, query_features, option["max_overlap"], option["max_overlap_percentage"]))
    if option["paths_limit"] > 1 and option["paths_limit"] < path_number:
        return None
    # PRIORITY CHECK: checking for number of instances - assess complexity of the feature graph
    if (len(query_features) > option["priority_threshold"] or path_number > option["max_cardinality"]) and \
            option["priority_mode"]:
        go_priority = True
        all_query_paths = "PRIORITY"
    elif option["priority_mode"]:
        # creating all paths
        all_query_paths = pb_graphtraversal(tmp_query_graph, [], [], option)
    elif len(tmp_query_graph) == 0:
        all_query_paths = []
    else:
        all_query_paths = "NOPRIORITY"
    return tmp_query_graph, all_query_paths, lin_query_set, query_features, a_q_f, query_clans, clan_dict, \
        go_priority, domain_count


def fc_main_sub(protein, domain_count, seed_proteome, option, all_query_paths, query_features, go_priority, a_q_f,
                clan_dict, tmp_query_graph, query_proteome, query, query_clans, domain_out, interprokeys, phmm):
    go_priority_2 = False
    mode = 0
    pathcount = 0
    weights = None
    unsolved = True
    if option["MS_uni"] == 0:
        if option["weight_const"]:
            weights, domain_count = w_weighting_constraints(protein, domain_count, seed_proteome, option)
        else:
            weights, domain_count = w_weighting(protein, domain_count, seed_proteome, option)
    search_protein, search_features, a_s_f = su_search_protein(protein, seed_proteome, option)
    search_protein = tuple(search_protein)
    max_fixture = ([], (0.0, 0.0, 0.0, 0.0, 0, 0, False), [], protein)
    region = pb_region_mapper(list(search_protein), search_features, option["max_overlap"],
                              option["max_overlap_percentage"])
    search_graph, search_path_number = pb_region_paths(region)
    # check for available paths
    # query <--VS-- seed
    # case M2.1: empty(query)

    if int(len(all_query_paths)) == 0 or int(len(query_features) == 0):
        # case M2.1.1: empty(query)-empty(search)
        # should be the best fix independent from weight
        if int(len(search_features)) == 0:
            path = list(a_s_f.keys())
            query_architecture = list(a_q_f.keys())
            score_w = sf_entire_calc_score(path, query_architecture, weights, search_features, a_s_f,
                                           query_features, a_q_f, clan_dict, option)
            mode = 0
        else:
            # case M2.1.2: empty(query)-graph(search)
            tmp_path_score = pb_entire_main_nongreedy(protein, [], search_features, weights, query_features,
                                                      seed_proteome, a_s_f, a_q_f, clan_dict, query_clans,
                                                      option, search_graph, search_path_number)
            path = tmp_path_score[0][0]
            score_w = tmp_path_score[0][1]
            query_architecture = tmp_path_score[0][2]
            mode = 0

        # set max_fixture according to
        max_fixture = (path, score_w, query_architecture, protein)
        unsolved = False

    # seed architecture == query architecture
    elif seed_proteome[protein] == query_proteome[query]:
        path = list(a_s_f.keys()) + list(search_features.keys())
        query_architecture = list(a_q_f.keys()) + list(query_features.keys())
        score_w = sf_entire_calc_score(path, query_architecture, weights, search_features, a_s_f,
                                       query_features, a_q_f, clan_dict, option)
        mode = 0
        max_fixture = (path, score_w, query_architecture, protein)
        unsolved = False

    # handle all paths
    # case M2.2: graph(query)
    elif not option["priority_mode"] and not int(len(query_features)) == 0:
        pool_results = None
        mode = 0
        stack, jobpaths = pb_create_jobs(tmp_query_graph, option)
        if len(stack) > 0:
            jobpool = multiprocessing.Pool(processes=option["cores"])
            func = partial(pb_graph_traversal_sub, protein, search_features, weights, query_features, seed_proteome,
                           a_s_f, a_q_f, clan_dict, query_clans, tmp_query_graph, option, search_graph,
                           search_path_number)
            try:
                pool_results = jobpool.map_async(func, stack).get(timeout=option["timelimit"])
            except multiprocessing.TimeoutError:
                go_priority_2 = True
                jobpool.terminate()
            if not go_priority_2:
                for pool_result in pool_results:
                    if pool_result[1][3] >= max_fixture[1][3]:
                        max_fixture = deepcopy(pool_result)
        if len(jobpaths) > 0 and not go_priority_2:
            # case M2.2.1: graph(query)-empty(search)
            if int(len(search_features)) == 0:
                for jobpath in jobpaths:
                    pathcount += 1
                    # special case: protein with no pfam or smart domains
                    # get score for a_s_f and query_path directly
                    path = list(a_s_f.keys())
                    query_path_ad = jobpath + list(a_q_f.keys())
                    score_w = sf_entire_calc_score(path, query_path_ad, weights, search_features, a_s_f,
                                                   query_features, a_q_f, clan_dict, option)

                    # check for max scoring fixture of path and query_path
                    if (score_w[4] >= max_fixture[1][4] and score_w[3] == max_fixture[1][3]) or \
                            score_w[3] >= max_fixture[1][3]:
                        max_fixture = (path, score_w, query_path_ad, protein)
            else:
                # case M2.2.2 graph(query)-graph(search)
                # regular traversal of graph based on search_protein
                jobpool = multiprocessing.Pool(processes=option["cores"])
                jobs = []
                jobcount = (int(len(jobpaths) / option["cores"]), len(jobpaths) % option["cores"])
                if jobcount[0] == 0:
                    for i in range(jobcount[1]):
                        jobs.append([jobpaths[-(i + 1)]])
                else:
                    counter = 0
                    for i in range(option["cores"]):
                        jobs.append(jobpaths[counter: (i + 1) * jobcount[0]])
                        counter = (i + 1) * jobcount[0]
                    for i in range(jobcount[1]):
                        jobs[i].append(jobpaths[-(i + 1)])
                func = partial(pb_calc_sub, protein, search_features, weights, query_features,
                               seed_proteome, a_s_f, a_q_f, clan_dict, query_clans, option, search_graph,
                               search_path_number)
                try:
                    pool_results = jobpool.map_async(func, jobs).get(timeout=option["timelimit"])
                except multiprocessing.TimeoutError:
                    go_priority_2 = True
                    jobpool.terminate()
                if not go_priority_2:
                    for pool_result in pool_results:
                        if pool_result[1][3] >= max_fixture[1][3]:
                            max_fixture = deepcopy(pool_result)
    if go_priority or go_priority_2 and unsolved:
        mode = 1
        all_query_paths = []
        priority_list = []
        for tool in option["input_linearized"]:
            for domain_type in query_proteome[query][tool]:
                priority_list.append(domain_type)
        priority_list.append("NONE")
        for domain_type in priority_list:
            all_query_paths += (
                pb_entire_graphtraversal_priority(tmp_query_graph, domain_type, protein, 1, search_features,
                                                  weights, query_features, seed_proteome, a_s_f, a_q_f,
                                                  clan_dict, query_clans, option))
            # max fixture of (search_path, score, query_path)
    if option["priority_mode"] or go_priority_2 and unsolved:
        for query_path in all_query_paths:
            pathcount += 1

            # case M2.2.1: graph(query)-empty(search)
            if int(len(search_features)) == 0:
                # special case: protein with no pfam or smart domains
                # get score for a_s_f and query_path directly
                path = list(a_s_f.keys())
                query_path_ad = query_path + list(a_q_f.keys())
                score_w = sf_entire_calc_score(path, query_path_ad, weights, search_features, a_s_f,
                                               query_features, a_q_f, clan_dict, option)
            else:
                # case M2.2.2 graph(query)-graph(search)
                # regular traversal of graph based on search_protein
                tmp_path_score = pb_entire_main_nongreedy(protein, query_path, search_features, weights,
                                                              query_features, seed_proteome, a_s_f, a_q_f, clan_dict,
                                                              query_clans, option, search_graph, search_path_number)

                path = tmp_path_score[0][0]
                score_w = tmp_path_score[0][1]
                query_path_ad = tmp_path_score[0][2]
                mode = tmp_path_score[1]

            # check for max scoring fixture of path and query_path
            if (score_w[4] >= max_fixture[1][4] and score_w[3] == max_fixture[1][3]) or \
                    score_w[3] >= max_fixture[1][3]:
                max_fixture = (path, score_w, query_path_ad, protein)

    score = max_fixture[1]
    try:
        if mode == 1:
            mode_out = "priority"
        else:
            mode_out = "exhaustive"
    except KeyError:
        mode_out = "default"
    if go_priority or go_priority_2:
        mode_out += "/priority"
    else:
        mode_out += "/exhaustive"

    best_template_path = []
    best_query_path = []

    for feature in max_fixture[0]:
        if feature in search_features:
            best_template_path.append((search_features[feature][0], search_features[feature][1],
                                       search_features[feature][2], search_features[feature][3]))
        else:
            best_template_path.append(
                (a_s_f[feature][0], a_s_f[feature][1], a_s_f[feature][2], a_s_f[feature][3]))

    for feature in max_fixture[2]:
        if feature in query_features:
            best_query_path.append((query_features[feature][0], query_features[feature][1],
                                    query_features[feature][2], query_features[feature][3]))
        else:
            best_query_path.append((a_q_f[feature][0], a_q_f[feature][1], a_q_f[feature][2], a_q_f[feature][3]))
    path_tmp = {}
    path_tmp_query = {}
    scale = 0
    if option["weight_const"] == 1:
        path_tmp2 = []
        for feature in best_template_path:
            if feature[0] not in path_tmp2:
                path_tmp2.append(feature[0])
        adjusted_weights = w_weight_const_rescale(path_tmp2, weights, search_features, True, option)
        for adj_feature in adjusted_weights:
            weights[adj_feature] = adjusted_weights[adj_feature]
    for feature in best_template_path:
        if feature[0] in path_tmp:
            path_tmp[feature[0]].append((feature[2], feature[3]))
        else:
            path_tmp[feature[0]] = [(feature[2], feature[3])]
            if option["MS_uni"] == 0:
                scale += weights[feature[0]]
    for feature in best_query_path:
        if feature[0] in path_tmp_query:
            path_tmp_query[feature[0]].append((feature[2], feature[3]))
        else:
            path_tmp_query[feature[0]] = [(feature[2], feature[3])]

    # unweighted case
    if option["MS_uni"] == 0:
        if scale > 0:
            scale = 1.0 / float(scale)
        else:
            scale = 1.0
    if option["domain"]:
        write_domain_out(seed_proteome, query_proteome, protein, query, weights, scale, path_tmp, path_tmp_query,
                             domain_out, option, interprokeys, phmm)
    if option["raw"]:
        print('#' + '\t' + protein + '\t' + query + '\t' + str(score[3]))
    return protein, query, (score[3], score[0], score[1], score[2], score[6]), mode_out


# Start Up Functions <su>


def su_lin_query_protein(protein_id, query_proteome, clan_dict, option):
    """Initializes variables for the current query protein (linearization version)

    :param option: dictionary that contains the main option variables of FAS
    :param protein_id: String that contains the identifier of the query protein
    :param query_proteome: dictionary that contains the feature architecture of all query proteins
    :param clan_dict: dictionary that maps features to clans
    :return:lin_query_protein, query_features, a_q_f(additional[not linearized] query features), query_clans, clan_dict
    """

    lin_query_protein = []
    query_clans = {}
    query_features = {}
    a_q_f = {}
    tmp = []
    i = 0

    for tool in option["input_linearized"]:
        for feature in query_proteome[protein_id][tool]:
            e_feature = False
            try:
                if query_proteome[protein_id][tool][feature]["evalue"] <= option["eFeature"]:
                    e_feature = True
            except TypeError:
                e_feature = True
            if e_feature:
                clan = None
                if feature in clan_dict:
                    clan = clan_dict[feature]
                tmp_clan = 0
                for instance in query_proteome[protein_id][tool][feature]["instance"]:
                    e_instance = False
                    try:
                        if instance[2] <= option["eInstance"]:
                            e_instance = True
                    except TypeError:
                        e_instance = True
                    if e_instance:
                        position = ((float(instance[0]) + float(instance[1])) / 2.0) / float(
                            query_proteome[protein_id]["length"])
                        position = round(position, 4)
                        key = "F_" + str(i)
                        query_features[key] = (feature, position, instance[0], instance[1])
                        tmp.append((key, instance[1]))
                        i += 1
                        tmp_clan += 1
                if clan in query_clans:
                    query_clans[clan] += tmp_clan
                elif tmp_clan > 0:
                    query_clans[clan] = tmp_clan
    for tool in option["input_normal"]:
        for feature in query_proteome[protein_id][tool]:
            e_feature = False
            try:
                if query_proteome[protein_id][tool][feature]["evalue"] <= option["eFeature"]:
                    e_feature = True
            except TypeError:
                e_feature = True
            if e_feature:
                clan = None
                if feature in clan_dict:
                    clan = clan_dict[feature]
                tmp_clan = 0
                for instance in query_proteome[protein_id][tool][feature]["instance"]:
                    e_instance = False
                    try:
                        if instance[2] <= option["eInstance"]:
                            e_instance = True
                    except TypeError:
                        e_instance = True
                    if e_instance:
                        position = ((float(instance[0]) + float(instance[1])) / 2.0) / float(
                            query_proteome[protein_id]["length"])
                        position = round(position, 4)
                        key = "O_" + str(i)
                        a_q_f[key] = (feature, position, instance[0], instance[1])
                        i += 1
                        tmp_clan += 1
                if clan in query_clans:
                    query_clans[clan] += tmp_clan
                elif tmp_clan > 0:
                    query_clans[clan] = tmp_clan

    # sort  instances
    tmp2 = sorted(tmp, key=itemgetter(1))
    for x in tmp2:
        lin_query_protein.append(x[0])

    return lin_query_protein, query_features, a_q_f, query_clans, clan_dict


def su_search_protein(protein_id, seed_proteome, option):
    """Initializes variables for the current seed protein

    :param option: dictionary that contains the main option variables of FAS
    :param protein_id: String that contains the identifier of the seed protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :return: search_protein, search_features, a_s_f (additional [not linearized] seed features)
    """
    search_features = {}
    search_protein = []
    a_s_f = {}
    tmp = []
    i = 0

    for tool in option["input_linearized"]:
        for feature in seed_proteome[protein_id][tool]:
            e_feature = False
            try:
                if seed_proteome[protein_id][tool][feature]["evalue"] <= option["eFeature"]:
                    e_feature = True
            except TypeError:
                e_feature = True
            if e_feature:
                for instance in seed_proteome[protein_id][tool][feature]["instance"]:
                    e_instance = False
                    try:
                        if instance[2] <= option["eInstance"]:
                            e_instance = True
                    except TypeError:
                        e_instance = True
                    if e_instance:
                        position = ((float(instance[0]) + float(instance[1])) / 2.0) / float(
                            seed_proteome[protein_id]["length"])
                        position = round(position, 4)
                        key = "F_" + str(i)
                        search_features[key] = (feature, position, instance[0], instance[1])
                        tmp.append((key, instance[1]))
                        i += 1
    for tool in option["input_normal"]:
        for feature in seed_proteome[protein_id][tool]:
            e_feature = False
            try:
                if seed_proteome[protein_id][tool][feature]["evalue"] <= option["eFeature"]:
                    e_feature = True
            except TypeError:
                e_feature = True
            if e_feature:
                for instance in seed_proteome[protein_id][tool][feature]["instance"]:
                    e_instance = False
                    try:
                        if instance[2] <= option["eInstance"]:
                            e_instance = True
                    except TypeError:
                        e_instance = True
                    if e_instance:
                        position = ((float(instance[0]) + float(instance[1])) / 2.0) / float(
                            seed_proteome[protein_id]["length"])
                        position = round(position, 4)
                        key = "O_" + str(i)
                        a_s_f[key] = (feature, position, instance[0], instance[1])
                        i += 1
    # sort  instances
    tmp2 = sorted(tmp, key=itemgetter(1))
    for x in tmp2:
        search_protein.append(x[0])
    return tuple(search_protein), search_features, a_s_f


# Path-building Functions # <pb>
# Used for the Pfam/Smart domains (default)
def pb_create_jobs(graph, option):
    """Does a breadth-first search to divide the feature architecture graph into several sub-problems so that it can be
    worked on by multiple cores. There are more jobs created than there are cores to counter the possibility of a core
    having nothing to do because of different running times for each job.

    :param graph: graph containing all paths through a protein architecture
    :param option: dictionary that contains the main option variables of FAS
    :return: queue (contains the starting points and sub-paths of the jobs), paths (empty unless b-f search reaches end
             node)
    """
    limit = 4 * option["cores"]
    queue = [("START", [])]
    paths = []
    while len(queue) < limit and queue:
        vertex, path = queue.pop(0)
        for next_vertex in graph[vertex]:
            if next_vertex == "END":
                paths.append(path)
            else:
                queue.append((next_vertex, path + [next_vertex]))
    return queue, paths


def pb_graph_traversal_sub(protein, search_features, weights, query_features, seed_proteome, a_s_f, a_q_f, clan_dict,
                           query_clans, query_graph, option, search_graph, search_path_number, stack):
    """Sub function for multi-processing: This function does a depth-first search in the feature graph. The starting
     point in the graph is given in the stack.

    :param search_path_number: number of paths in search protein
    :param search_graph: architecture graph of the search protein
    :param protein: String that contains the identifier of the seed protein
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :param stack: contains the starting point in the graph and sub-path
    :param query_graph: graph of the query architecture
    :return: max_fixture
    """
    max_fixture = ([], (0.0, 0.0, 0.0, 0.0, 0, 0, False), [], protein)
    v_stack, p_stack = [stack[0]], [stack[1]]
    while v_stack:
        query_path, v_stack, p_stack = pb_graphtraversal(query_graph, v_stack, p_stack, option)
        # case M2.2.1: graph(query)-empty(search)
        if int(len(search_features)) == 0:
            # special case: protein with no pfam or smart domains
            # get score for a_s_f and query_path directly
            path = list(a_s_f.keys())
            query_path_ad = query_path + list(a_q_f.keys())
            score_w = sf_entire_calc_score(path, query_path_ad, weights, search_features, a_s_f,
                                           query_features, a_q_f, clan_dict, option)
        else:
            # case M2.2.2 graph(query)-graph(search)
            # regular traversal of graph based on search_protein
            tmp_path_score = pb_entire_main_nongreedy(protein, query_path, search_features, weights, query_features,
                                                      seed_proteome, a_s_f, a_q_f, clan_dict, query_clans, option,
                                                      search_graph, search_path_number)
            path = tmp_path_score[0][0]
            score_w = tmp_path_score[0][1]
            query_path_ad = tmp_path_score[0][2]

        # check for max scoring fixture of path and query_path
        if ((not max_fixture[1][5]) and (score_w[4] >= max_fixture[1][4]) and score_w[3] <= max_fixture[1][3]) or \
                score_w[3] >= max_fixture[1][3]:
            max_fixture = (path, score_w, query_path_ad, protein)
    return max_fixture


def pb_calc_sub(protein, search_features, weights, query_features, seed_proteome, a_s_f, a_q_f, clan_dict, query_clans,
                option, search_graph, search_path_number, jobpaths):
    """Sub function for multiprocessing [2]: Goes through a list of (query) paths an evaluates them.

    :param protein: String that contains the identifier of the seed protein
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :param jobpaths: a group of (query) paths to be evaluated
    :param search_path_number: number of paths in search protein
    :param search_graph: architecture graph of the search protein
    :return: max_fixture, tmpcheck - tmpstart (time taken for calculation)

    """
    max_fixture = ([], (0.0, 0.0, 0.0, 0.0, 0, 0, False), [], protein)
    for jobpath in jobpaths:
        tmp_path_score = pb_entire_main_nongreedy(protein, jobpath, search_features, weights, query_features,
                                                  seed_proteome, a_s_f, a_q_f, clan_dict, query_clans, option,
                                                  search_graph, search_path_number)
        path = tmp_path_score[0][0]
        score_w = tmp_path_score[0][1]
        query_path_ad = tmp_path_score[0][2]
        if ((not max_fixture[1][5]) and (score_w[4] >= max_fixture[1][4]) and score_w[3] <= max_fixture[1][3]) or \
                score_w[3] >= max_fixture[1][3]:
            max_fixture = (path, score_w, query_path_ad, protein)
    return max_fixture


def pb_entire_main_nongreedy(protein_id, query_path, search_features, weights, query_features, seed_proteome, a_s_f,
                             a_q_f, clan_dict, query_clans, option, search_graph, search_path_number):
    """Main Path-building function,
    creates graph with all paths(for seed/search protein), checks priority mode activation if necessary retrieves best
    path from graph traversal function, returns best path, score and mode
    Function calls: pb_region_mapper(), pb_region_paths(), pb_entire_priority_mode(),
                    pb_entire_graphtraversal()


    :param protein_id: String that contains the identifier of the seed protein
    :param query_path: currently evaluated query path
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :param search_path_number: number of paths in search protein
    :param search_graph: architecture graph of the search protein
    :param search_path_number: number of paths in search protein
    :param search_graph: architecture graph of the search protein
    :return: path_score, mode[priority or extensive]

    """
    if (int(len(search_features)) >= option["priority_threshold"] or search_path_number > option['max_cardinality']) \
            and option["priority_mode"]:
        mode = 1
        path_score = pb_entire_priority_mode(protein_id, query_path, search_graph, search_features, weights,
                                             query_features, seed_proteome, a_s_f, a_q_f, clan_dict, query_clans,
                                             option)
    else:
        mode = 0
        path_score = pb_entire_graphtraversal(search_graph, query_path, search_features, weights, query_features,
                                              a_s_f, a_q_f, clan_dict, option)
    return path_score, mode


def pb_entire_priority_mode(protein, query_path, search_graph, search_features, weights, query_features, seed_proteome,
                            a_s_f, a_q_f, clan_dict, query_clans, option):
    """Path-evaluation in priority mode

    :param protein: String that contains the identifier of the seed protein
    :param query_path: currently evaluated query path
    :param search_graph: graph containing all paths through the seed architecture
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :return: best_path
    """

    # best_path (Path, (MS_score(float), PS_score(float), CS_score(float), final_score(float), path_weight(float),
    # common_feature(bool)), QPath)
    best_path = ("NULL", (0.0, 0.0, 0.0, 0.0, 0.0, False), "NULL")
    priority_list = []
    for tool in option["input_linearized"]:
        for domain_type in seed_proteome[protein][tool]:
            priority_list.append(domain_type)
    priority_list.append("NONE")
    for domain_type in priority_list:
        path_score_w = pb_entire_graphtraversal_priority(search_graph, domain_type, query_path, 0, search_features,
                                                         weights, query_features, seed_proteome, a_s_f, a_q_f,
                                                         clan_dict, query_clans, option)
        if path_score_w[1][5]:
            if path_score_w[1][3] >= best_path[1][3]:
                best_path = (path_score_w[0], path_score_w[1], path_score_w[2])
        else:
            # if so far no path with common features found AND the weight is higher
            if (not best_path[1][5]) and (path_score_w[1][4] >= best_path[1][4]):
                best_path = (path_score_w[0], path_score_w[1], path_score_w[2])
    return best_path


def pb_entire_graphtraversal(search_graph, query_path, search_features, weights, query_features, a_s_f, a_q_f,
                             clan_dict, option):
    """Traverses the feature architecture graph (seed) to score all paths

    :param search_graph: graph containing all paths through the seed architecture
    :param query_path: currently evaluated query path
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param option: dictionary that contains the main option variables of FAS
    :return: best_path
    """

    v_stack = ["START"]
    p_stack = []
    best_path = ([], (0.0, 0.0, 0.0, 0.0, 0.0, False, 0.0), [])
    first = 1
    if len(query_path) > 0:
        while v_stack:
            vertex = v_stack.pop()
            if len(p_stack) == 0:
                path = []
            else:
                path = p_stack.pop()
            for next_vertex in search_graph[vertex]:
                if next_vertex == "END":
                    path_ad = path + list(a_s_f.keys())
                    query_path_ad = query_path + list(a_q_f.keys())
                    score_w = sf_entire_calc_score(path_ad, query_path_ad, weights, search_features, a_s_f,
                                                   query_features, a_q_f, clan_dict, option)
                    if (score_w[4] >= best_path[1][4] and score_w[3] == best_path[1][3]) or \
                       score_w[3] >= best_path[1][3]:
                        best_path = (path_ad, score_w, query_path_ad)
                else:
                    v_stack.append(next_vertex)
                    p_stack.append(path + [next_vertex])
    else:
        while v_stack and first:
            vertex = v_stack.pop()
            if len(p_stack) == 0:
                path = []
            else:
                path = p_stack.pop()
            for next_vertex in search_graph[vertex]:
                if next_vertex == "END":
                    path_ad = path + list(a_s_f.keys())
                    query_path_ad = query_path + list(a_q_f.keys())
                    score_w = sf_entire_calc_score(path_ad, query_path_ad, weights, search_features, a_s_f,
                                                   query_features, a_q_f, clan_dict, option)
                    if (score_w[4] >= best_path[1][4] and score_w[3] == best_path[1][3]) or \
                       score_w[3] >= best_path[1][3]:
                        best_path = (path_ad, score_w, query_path_ad)
                    first = 0
                else:
                    v_stack.append(next_vertex)
                    p_stack.append(path + [next_vertex])
    return best_path


def pb_graphtraversal(graph, v_stack, p_stack, option):
    """Traverses the feature architecture graph (query) to evaluate number of paths, returns all paths if # of paths is
     smaller than the max_cardinality

    :param graph: graph containing all paths through the (query) architecture
    :param option: dictionary that contains the main option variables of FAS
    :param v_stack: vertex stack
    :param p_stack: path stack
    :return: paths
    """
    paths = []
    if v_stack == [1]:
        v_stack = ["START"]
        p_stack = []
        mode = 1
    elif not v_stack:
        v_stack = ["START"]
        p_stack = []
        mode = 0
    else:
        mode = 1
    while v_stack:
        vertex = v_stack.pop()
        if len(p_stack) == 0:
            path = []
        else:
            path = p_stack.pop()
        for next_vertex in graph[vertex]:
            if next_vertex == "END":
                if mode == 0:
                    paths.append(path)
                    if len(paths) > option["max_cardinality"]:
                        return paths
                else:
                    return path, v_stack, p_stack
            else:
                v_stack.append(next_vertex)
                p_stack.append(path + [next_vertex])
    return paths


def pb_entire_graphtraversal_priority(search_graph, priority, query_path, mode, search_features, weights,
                                      query_features, seed_proteome, a_s_f, a_q_f, clan_dict, query_clans,
                                      option):
    """Traverses the feature architecture graph in priority mode

    :param search_graph: graph containing all paths through the seed architecture
    :param priority: current priority feature
    :param query_path: graph containing all paths through the query architecture
    :param mode: query or seed linearization
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :return: paths or best_path
    """

    v_stack = ["START"]
    p_stack = []
    protein = None
    paths = []
    best_path = None
    if mode == 0:
        best_path = ([], (0.0, 0.0, 0.0, 0.0, 0.0, False), [])
    else:
        protein = query_path
    while v_stack:
        vertex = v_stack.pop()
        if len(p_stack) == 0:
            path = []
        else:
            path = p_stack.pop()

        p_found = 0
        p_candidates = []
        for next_vertex in search_graph[vertex]:
            if next_vertex == "END":
                if mode == 1:
                    paths.append(path)
                    p_found = 2
                else:
                    path_ad = path + list(a_s_f.keys())
                    query_path_ad = query_path + list(a_q_f.keys())
                    score_w = sf_entire_calc_score(path_ad, query_path_ad, weights, search_features, a_s_f,
                                                   query_features, a_q_f, clan_dict, option)
                    if ((score_w[4] >= best_path[1][4] and score_w[3] == best_path[1][3])
                            or score_w[3] >= best_path[1][3]):
                        best_path = (path_ad, score_w, query_path_ad)
                    p_found = 2
            elif mode == 1:
                if query_features[next_vertex][0] == priority:
                    p_candidates.append(next_vertex)
                    p_found = 1
            elif mode == 0:
                p_candidates.append(next_vertex)
                p_found = 1
        if p_found == 1:
            if len(p_candidates) == 1:
                v_stack.append(p_candidates[0])
                p_stack.append(path + p_candidates)
            else:
                if mode == 1:
                    best_priority_bridger = ("NONE", (0.0, 0.0, 0.0, 0.0, 0.0, False))
                    for next_vertex in p_candidates:
                        score = sf_calc_score(path + [next_vertex], protein, weights, search_features, query_features,
                                              seed_proteome, clan_dict, query_clans, option)
                        if score[3] >= best_priority_bridger[1][3]:
                            best_priority_bridger = (next_vertex, score)
                    v_stack.append(best_priority_bridger[0])
                    p_stack.append(path + [best_priority_bridger[0]])
                elif mode == 0:
                    # greedy strategy: if feature type priority appears more than once
                    best_partial_path = ("NONE", (0.0, 0.0, 0.0, 0.0, 0.0, False))
                    for next_vertex in p_candidates:
                        path_ad = path + list(a_s_f.keys())
                        query_path_ad = query_path + list(a_q_f.keys())
                        score_w = sf_entire_calc_score(path_ad + [next_vertex], query_path_ad, weights,
                                                       search_features, a_s_f, query_features, a_q_f, clan_dict, option)
                        if (score_w[4] >= best_partial_path[1][4] and score_w[3] == best_partial_path[1][3]) \
                           or score_w[3] >= best_partial_path[1][3]:
                            best_partial_path = (next_vertex, score_w)

                    v_stack.append(best_partial_path[0])
                    p_stack.append(path + [best_partial_path[0]])

        elif p_found == 0:
            if mode == 1:
                best_priority_bridger = ("NONE", (0.0, 0.0, 0.0, 0.0, 0.0, False))
                for next_vertex in search_graph[vertex]:
                    score = sf_calc_score(path + [next_vertex], protein, weights, search_features, query_features,
                                          seed_proteome, clan_dict, query_clans, option)
                    if score[3] >= best_priority_bridger[1][3]:
                        best_priority_bridger = (next_vertex, score)
                v_stack.append(best_priority_bridger[0])
                p_stack.append(path + [best_priority_bridger[0]])

            elif mode == 0:
                # greedy strategy: if feature type (p priority) not found
                best_partial_path = ("NONE", (0.0, 0.0, 0.0, 0.0, 0.0, False))
                for next_vertex in search_graph[vertex]:
                    path_ad = path + list(a_s_f.keys())
                    query_path_ad = query_path + list(a_s_f.keys())
                    score_w = sf_entire_calc_score(path_ad + [next_vertex], query_path_ad, weights, search_features,
                                                   a_s_f, query_features, a_q_f, clan_dict, option)
                    if (score_w[4] >= best_partial_path[1][4] and score_w[3] == best_partial_path[1][3]) or \
                       score_w[3] >= best_partial_path[1][3]:
                        best_partial_path = (next_vertex, score_w)

                v_stack.append(best_partial_path[0])
                p_stack.append(path + [best_partial_path[0]])
    if mode == 1:
        return paths
    elif mode == 0:
        return best_path



#!/bin/env python

#######################################################################
# Copyright (C) 2023 Vinh Tran
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


import os
import sys
import argparse
import json
import glob
from pkg_resources import get_distribution


def merge_JsonFiles(json_files, outName, outPath):
    out_dict = {}
    # Load each JSON file
    for json_file in json_files:
        with open(json_file, "r") as f:
            dd = json.load(f)
            for id in list(dd):
                tmp = id.split('_')
                if not len(tmp) == 2:
                    sys.exit(f'{json_file} seems not to be a valid FAS output!')
                if f'{tmp[0]}_{tmp[1]}' in out_dict or f'{tmp[1]}_{tmp[0]}' in out_dict:
                    del dd[id]
            out_dict.update(dd)
    # Dump into a single JSON file
    final_out = f'{os.path.abspath(outPath)}/{outName}.json'
    if not os.path.exists(final_out):
        with open(final_out, "w") as f:
            json.dump(out_dict, f)
    else:
        sys.exit(f'Output file {final_out} exists!')


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.' +
                                     ' This script is used to merge multiple calculated FAS json output files!' +
                                     ' Note: if one protein pair has different scores in different files, only' +
                                     ' the scores in the first file will be kept. If you want to update the FAS scores,' +
                                     ' please use fas.updateJson function!',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument("-i", "--input", default='.', type=str, required=True,
                          help="Path to input folder contatining json files")
    required.add_argument("-n", "--outName", default=None, type=str, required=True,
                          help="Name of the output json")
    optional.add_argument("-o", "--outPath", default='.', type=str, required=False,
                          help="Path to output directory")
    args = parser.parse_args()

    path = f'{os.path.abspath(args.input)}/*.json'
    files = glob.glob(path)
    merge_JsonFiles(files, args.outName, args.outPath)
    print(f'DONE! Final output saved in {os.path.abspath(args.outPath)}/{args.outName}.json')


if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2023 Vinh Tran
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import os
import sys
import argparse
import json
from pkg_resources import get_distribution
from greedyFAS.annoFAS import annoModules


def update_annoFile(js_old, js_new, force):
    if os.path.exists(f'{js_old}.updated'):
        if not force:
            sys.exit(f'ERROR: Output file {js_old}.updated exists!')
    # Load each JSON file
    with open(js_new, 'r') as f1:
        anno_new = json.load(f1)
    with open(js_old, 'r') as f2:
        anno_old = json.load(f2)
        for i in anno_old:
            anno_old[i].update(anno_new[i])
    # Dump into a single JSON file
    with open(f'{js_old}.updated', "w") as f:
        json.dump(anno_old, f)


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.' +
                                     ' This script is used to update an old annotation json file by a new one!',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument("--old", default='.', type=str, required=True,
                          help="Old annotation json file")
    required.add_argument("--new", default=None, type=str, required=True,
                          help="New annotation json file")
    optional.add_argument('--force', help='Force override output files', action='store_true')
    args = parser.parse_args()

    annoModules.checkFileExist(args.new)
    annoModules.checkFileExist(args.old)

    update_annoFile(args.old, args.new, args.force)
    print(f'DONE! Final output saved in {os.path.abspath(args.old)}.updated')


if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2020 Julian Dosch
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


import os
import sys
import argparse
import multiprocessing as mp
from sys import argv
from greedyFAS.annoFAS import annoFAS
from greedyFAS.annoFAS import annoModules
from greedyFAS.mainFAS import fasInput, fasOutput, greedyFAS
from greedyFAS.mainFAS.fasInput import read_json
from pkg_resources import get_distribution


def get_options():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    general = parser.add_argument_group('general arguments')
    inargs = parser.add_argument_group('input arguments')
    outargs = parser.add_argument_group('output arguments')
    annotation = parser.add_argument_group('annotation arguments')
    weighting = parser.add_argument_group('weighting arguments')
    thresholds = parser.add_argument_group('threshold arguments')
    obscure = parser.add_argument_group('obscure arguments')
    parser.add_argument('--version', action='version', version=str(version))
    required.add_argument("-s", "--seed", default=None, type=str, required=True,
                          help="path to seed protein fasta file")
    required.add_argument("-q", "--query", default=None, type=str, required=True,
                          help="path to query protein fasta file")
    required.add_argument("-a", "--annotation_dir", default=None, type=str, required=True,
                          help='working directory, all annotations are be stored here')
    required.add_argument("-o", "--out_dir", default=None, type=str, required=True,
                          help="output directory, all outputfiles will be stored here")
    general.add_argument("--bidirectional", action="store_true",
                         help="calculate both scoring directions")
    general.add_argument('--cpus', help='number of cores', type=int, action='store', default=0)
    annotation.add_argument('--forceAnno', help='Force override annotations', action='store_true')
    annotation.add_argument("-f", "--eFeature", default="0.001", type=float,
                            help="eValue cutoff for PFAM/SMART domain, applied during annotation but also during "
                                 "calculation")
    annotation.add_argument("-i", "--eInstance", default="0.01", type=float,
                            help="eValue cutoff for PFAM/SMART instances, applied during annotation but also during "
                                 "calculation")
    annotation.add_argument('--eFlps', help='eValue cutoff for fLPS', action='store', default=0.0000001, type=float)
    annotation.add_argument('--org', help='Organism of input sequence(s) for SignalP search',
                            choices=['euk', 'gram+', 'gram-'], action='store', default='euk', type=str)
    annotation.add_argument("--toolPath", dest="toolPath", default='', type=str,
                            help="Path to Annotion tool directory created with fas.setup")
    weighting.add_argument("-r", "--ref_proteome", default=None, type=str,
                           help="Path to a reference proteome which can be used for the weighting of features, "
                                "by default there is no reference proteome used, the weighting will be uniform")
    weighting.add_argument("--ref_2", dest="ref_2", default=None, type=str,
                           help="Give a second reference for bidirectional mode, does not do anything if bidirectional "
                                "mode is not active or if no main reference was given")
    weighting.add_argument("-g", "--weight_correction", default="loge", type=str,
                           help="Function applied to the frequency of feature types during weighting, options are "
                                "linear(no function), loge(natural logarithm[Default]), log10(base-10 logarithm), "
                                "root4(4th root) and root8(8th root).")
    weighting.add_argument("-x", "--weight_constraints", default=None, type=str,
                           help="Apply weight constraints via constraints file, by default there are no constraints. "
                                "Please look at the FAS wiki pages for templates for the constraints file")
    inargs.add_argument("--query_id", default=None, nargs='*', type=str,
                        help="Choose specific proteins (ids divided by spaces) from the query input for calculation, "
                             "by default this is off (all proteins are used for calculation)")
    inargs.add_argument("--seed_id", default=None, nargs='*', type=str,
                        help="Choose specific proteins (ids divided by spaces) from the seed input for calculation, "
                             "by default this is off (all proteins are used for calculation)")
    inargs.add_argument("--pairwise", dest="pairwise", default=None, type=str,
                        help="deactivate all against all comparison, needs a pairing file with the ids that should be"
                             " compared (one pair per line tab seperated), please look at the FAS wiki pages for "
                             "templates")
    inargs.add_argument("--oldJson", default='', type=str,
                        help="Input old FAS output in JSON format. Only new pairs will be considered!")
    inargs.add_argument("-d", "--featuretypes", default='', type=str,
                        help="inputfile that contains the tools/databases used to predict features. Please look at the "
                             "FAS wiki pages for templates of the the featuretypes input file")
    inargs.add_argument("--extra_annotation", default=None, nargs='*', type=str,
                        help="give naming conventions for extra annotation files, eg. if the file name for the seed is "
                             "someseed.json than the extra annotation files should be named someseed_"
                             "[EXTRA_ANNOTATION].json, these extra files need to exist for seed, query and the "
                             "references (if given)")
    outargs.add_argument("-n", "--out_name", default=None, type=str,
                         help="name for outputfiles, if none is given the name will be created from the seed and "
                              "query names")
    outargs.add_argument("--raw", dest="raw", action="store_true",
                         help="print FAS scores to terminal")
    outargs.add_argument("--tsv", dest="tsv", action="store_true",
                         help="deactivates creation of the tsv output, either use together with --raw or "
                              "--phyloprofile")
    outargs.add_argument("--json", dest="json", action="store_true",
                         help="create json output")
    outargs.add_argument("--phyloprofile", dest="phyloprofile", default=None, type=str,
                         help="activate phyloprofile output, needs mapping file for all query proteins")
    outargs.add_argument("--domain", dest="domain", action="store_false",
                         help="deactivate .domains output")
    outargs.add_argument("--no_config", dest="config", action="store_false",
                         help="deactivate *_config.yml output")
    outargs.add_argument('--force', help='Force override output files', action='store_true')
    outargs.add_argument('--silent', help='Turn off STDOUT', action='store_true')
    thresholds.add_argument("-c", "--max_overlap", dest="max_overlap", default=0, type=int,
                            help="maximum size overlap allowed, default is 0 amino acids")
    thresholds.add_argument("--max_overlap_percentage", dest="max_overlap_percentage", default=0.4, type=float,
                            help="defines how much percent of a feature the overlap is allowed to cover, default "
                                 "is 0.4 (40%%)")
    thresholds.add_argument("-t", "--priority_threshold", type=int, default=30,
                            help="Change to define the feature number threshold for activating priority mode in the "
                                 "path evaluation. default=30")
    thresholds.add_argument("-m", "--max_cardinality", default=500, type=int,
                            help="Change to define the threshold for the maximal cardinality (number) of feature paths "
                                 "in a graph. If max. cardinality is exceeded the priority mode will be used to for "
                                 "the path evaluation. default=500")
    thresholds.add_argument("--paths_limit", default=0, type=int,
                            help="Specify number of maximum paths to be considered (10^n). If this threshold is exceeded, "
                                "the corresponding protein will be ignored. Default: 0 for no limit")
    obscure.add_argument("--priority_mode", action='store_false',
                         help="deactivates the greedy strategy priority mode for larger architectures, NOT RECOMMENDED")
    obscure.add_argument("--timelimit", default=3600, type=int,
                         help="Sets a soft time limit in seconds for the calculation between a pair of proteins,"
                              "This limit does not necessarily represent the actual runtime. It only stops the "
                              "exhaustive strategy through the architecture and activate priority mode. "
                              "This option is only relevant if the option [--priority_mode] is set")
    obscure.add_argument("-w", "--score_weights", nargs=3, default=[0.7, 0.0, 0.3], type=float,
                         help="Defines how the three scores MS, CS and PS are weighted, takes three float arguments, "
                              "sum should be 1.0, the default is 0.7, 0.0, 0.3")
    obscure.add_argument("--empty_as_1", action='store_true',
                         help="If both proteins have no features, score 1.0 instead of the default 0.0")
    args = parser.parse_args()
    return args


def anno(annojobs, args, toolpath):
    eflps = args.eFlps
    signalporg = args.org
    efeature = args.eFeature
    einstance = args.eInstance
    hmmcores = 1
    cpus = args.cpus
    if cpus == 0:
        cpus = mp.cpu_count()-1
    for annojob in annojobs:
        name = '.'.join(annojob.split('/')[-1].split('.')[:-1])
        outpath = os.path.abspath(args.annotation_dir)
        annotate = True
        seqfile = annojob
        if not args.silent:
            print('Check annotation for "' + name + '"...')
        if os.path.exists(os.path.abspath(args.annotation_dir + '/' + name + '.json')):
            if not args.silent:
                print('Annotation for "' + name + '" already exists.')
            if args.forceAnno:
                if not args.silent:
                    print('Overwriting...')
                annoModules.checkFileExist(seqfile)
            else:
                annotate = False
        else:
            annoModules.checkFileExist(seqfile)
        if annotate:
            annoFAS.runAnnoFas(
                [seqfile, outpath, toolpath, args.forceAnno, name, eflps, signalporg, efeature, einstance, hmmcores, '',
                 '', '', cpus, args.featuretypes])

def fas(opts):
    (args, toolpath) = opts
    option_dict = {
                   "weight_const": False, "seed_id": args.seed_id, "query_id": args.query_id,
                   "priority_mode": args.priority_mode, "priority_threshold": args.priority_threshold,
                   "max_cardinality": args.max_cardinality, "cores": int(args.cpus), "raw": args.raw,
                   "bidirectional": args.bidirectional, "max_overlap": args.max_overlap,
                   "timelimit": args.timelimit, "phyloprofile": args.phyloprofile, "score_weights": [],
                    "tsv": args.tsv, "json": args.json, "max_overlap_percentage": 0.0, "domain": args.domain, "pairwise": None,
                    "eInstance": args.eInstance, "eFeature": args.eFeature, "progress": True,
                    "empty_as_1": args.empty_as_1, "silent": args.silent, "paths_limit": 10**args.paths_limit
                   }
    seedname = '.'.join(args.seed.split('/')[-1].split('.')[:-1])
    option_dict["p_path"] = [args.annotation_dir + '/' + seedname + '.json']
    queryname = '.'.join(args.query.split('/')[-1].split('.')[:-1])
    option_dict["s_path"] = [args.annotation_dir + '/' + queryname + '.json']
    if args.ref_proteome:
        name = '.'.join(args.ref_proteome.split('/')[-1].split('.')[:-1])
        option_dict["ref_proteome"] = [args.annotation_dir + '/' + name + '.json']
    else:
        option_dict["ref_proteome"] = None
    if args.raw:
        option_dict["progress"] = False
    if args.silent:
        option_dict["progress"] = False
    if option_dict["cores"] == 0:
        option_dict["cores"] = mp.cpu_count()-1
    if args.ref_2:
        r2name = '.'.join(args.ref_2.split('/')[-1].split('.')[:-1])
        option_dict["ref_2"] = [args.annotation_dir + '/' + r2name + '.json']
    else:
        option_dict["ref_2"] = None
    if args.extra_annotation:
        for i in args.extra_annotation:
            option_dict["p_path"].append(args.annotation_dir + '/' + i + '.json')
            option_dict["s_path"].append(args.annotation_dir + '/' + i + '.json')
            if args.ref_proteome:
                option_dict["ref_proteome"].append(args.annotation_dir + '/' + i + '.json')
            if args.ref_2:
                option_dict["ref_2"].append(args.annotation_dir + '/' + i + '.json')
    if not option_dict["ref_proteome"]:
        option_dict["MS_uni"] = 1
        option_dict["weight_correction"] = None
    try:
        test = 0.0
        for weight in args.score_weights:
            option_dict["score_weights"].append(weight)
            test += float(weight)
        if test != 1.0:
            raise Exception("The sum of all values in -w [--score_weights] should be 1.0")
        option_dict["score_weights"] = tuple(option_dict["score_weights"])
    except ValueError:
        raise Exception(str(args.score_weights) + " is not a valid input for -w [--score_weights]")
    if not option_dict["ref_proteome"]:
        option_dict["MS_uni"] = 1
        option_dict["weight_correction"] = None
    else:
        option_dict["MS_uni"] = 0
        if args.weight_correction == "log10":
            option_dict["weight_correction"] = "log10"
        elif args.weight_correction == "loge":
            option_dict["weight_correction"] = "loge"
        elif args.weight_correction == "root4":
            option_dict["weight_correction"] = "root4"
        elif args.weight_correction == "root8":
            option_dict["weight_correction"] = "root8"
        elif args.weight_correction == "linear":
            option_dict["weight_correction"] = None
        else:
            raise Exception(str(args.weightcorrection) + "is not a valid input for -g [--weightcorrection]")
    if args.ref_proteome and args.weight_constraints:
        option_dict["weight_const"] = True
        option_dict["constraints"] = fasInput.constraints_in(args.weight_constraints)
    elif args.weight_constraints:
        raise Exception("[--weight_constraints] only works with a reference proteome")
    if 0.0 <= float(args.max_overlap_percentage) <= 1.0:
        option_dict["max_overlap_percentage"] = float(args.max_overlap_percentage)
    else:
        raise Exception("[--max_overlap_percentage] should be between 0.0 and 1.0")
    if args.out_name:
        option_dict['outpath'] = args.out_dir.rstrip('/') + '/' + args.out_name
        out_file = f'{option_dict["outpath"]}.tsv'
        if args.json:
            out_file = f'{option_dict["outpath"]}.json'
        if os.path.exists(out_file):
            if not args.force:
                sys.exit(f'Output file {os.path.abspath(out_file)} exists! Use --force if you want to overwrite it!')
            else:
                os.remove(f'{os.path.abspath(out_file)}')
    else:
        option_dict['outpath'] = args.out_dir.rstrip('/') + '/' + seedname + '_' + queryname
    if args.featuretypes:
        option_dict['input_linearized'], option_dict['input_normal'] = fasInput.featuretypes(args.featuretypes)
    else:
        option_dict['input_linearized'], option_dict['input_normal'] = fasInput.featuretypes(toolpath
                                                                                             + '/' + 'annoTools.txt')

    option_dict["old_json"] = False
    if args.oldJson:
        if os.path.exists(os.path.abspath(args.oldJson)):
            if not args.silent:
                print(f'### NOTE: existing output given ({os.path.abspath(args.oldJson)}). Only new pairs of proteins will be considered!')
            option_dict["old_json"] = os.path.abspath(args.oldJson)
        else:
            print(f'WARNING: {args.oldJson} not found!')

    if args.pairwise:
        option_dict["pairwise"] = fasInput.read_pairwise(args.pairwise)
        if args.oldJson:
            if os.path.exists(os.path.abspath(args.oldJson)):
                option_dict["pairwise"] = filter_oldJson(args.oldJson, fasInput.read_pairwise(args.pairwise))
        if len(option_dict['pairwise']) == 0:
            sys.exit(f'DONE! All pairwise FAS scores can be found in {args.oldJson}!')
    else:
        option_dict["pairwise"] = None

    if not args.silent:
        print('Calculating FAS score...')
    greedyFAS.fc_start(option_dict)
    if args.config:
        fasOutput.write_metadata(option_dict['outpath'] + '_config.yml', args, ' '.join(argv),
                                 str(get_distribution('greedyFAS').version))
    if not args.silent:
        print('done!')


def filter_oldJson(old_json, in_pairs):
    """ Function to check if a FAS scores for input protein pair already exists
    Return a list of new pairs only
    """
    old_results = read_json(old_json)
    new_pairs = []
    for pair in in_pairs:
        if not f'{pair[0]}_{pair[1]}' in old_results and not f'{pair[1]}_{pair[0]}' in old_results:
            new_pairs.append(pair)
    return(new_pairs)


def main():
    args = get_options()
    toolpath = args.toolPath
    if toolpath == '':
        pathconfigfile = os.path.realpath(__file__).replace('calcFAS.py', 'pathconfig.txt')
        with open(pathconfigfile) as f:
            toolpath = f.readline().strip()
    else:
        toolpath = os.path.abspath(args.toolPath)
    if not os.path.isdir(args.annotation_dir):
        os.mkdir(args.annotation_dir)
    if not os.path.isdir(args.out_dir):
        os.mkdir(args.out_dir)
    annojobs = [args.seed]
    if args.query not in annojobs:
        annojobs.append(args.query)
    if args.ref_proteome and args.ref_proteome not in annojobs:
        annojobs.append(args.ref_proteome)
    if args.ref_2 and args.ref_2 not in annojobs:
        annojobs.append(args.ref_2)
    anno(annojobs, args, toolpath)
    fas((args, toolpath))


if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2023 Vinh Tran
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


import os
import sys
import argparse
import json
import glob
from pkg_resources import get_distribution
from greedyFAS.annoFAS import annoModules
from pkg_resources import get_distribution


def update_JsonFile(js_old, js_new, out_file, force):
    if os.path.exists(f'{js_old}.updated'):
        if not force:
            sys.exit(f'ERROR: Output file {js_old}.updated exists!')
    # Load each JSON file
    with open(js_new, 'r') as f1:
        fas_new = json.load(f1)
        print(f'Pairs in new file: {len(fas_new)}')
    with open(js_old, 'r') as f2:
        fas_old = json.load(f2)
        print(f'Pairs in old file: {len(fas_old)}')
        c = 0
        for id in list(fas_old):
            tmp = id.split('_')
            if not len(tmp) == 2:
                sys.exit(f'{js_old} seems not to be a valid FAS output!')
            if f'{tmp[0]}_{tmp[1]}' in fas_new or f'{tmp[1]}_{tmp[0]}' in fas_new:
                c += 1
                del fas_old[id]
        fas_old.update(fas_new)
        print(f'Updated pairs: {c}')
        print(f'Total pairs in final file: {len(fas_old)}')
    # Dump into a single JSON file
    if not out_file:
        out_file = f'{js_old}.updated'
    with open(out_file, "w") as f:
        json.dump(fas_old, f)
    return(out_file)


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.' +
                                     ' This script is used to update an old FAS json file by a new one!',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument("--old", default='.', type=str, required=True,
                          help="Old FAS json file")
    required.add_argument("--new", default=None, type=str, required=True,
                          help="New FAS json file")
    optional.add_argument("--out", default=None, type=str,
                          help="Output file")
    optional.add_argument('--force', help='Force override output files', action='store_true')
    args = parser.parse_args()

    annoModules.checkFileExist(args.new)
    annoModules.checkFileExist(args.old)

    out_file = update_JsonFile(args.old, args.new, args.out, args.force)
    print(f'DONE! Final output saved in {out_file}')


if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2023 Vinh Tran
#
#  This file is part of FAS. This script is used to convert FAS output
#  from TSV (tab-delimited) to JSON formar. Only proteins IDs and their
#  (pairwise) FAS scores will be stored in the JSON output.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


import os
import sys
import argparse
import json
import glob
from pkg_resources import get_distribution


def read_file(file):
    """ Read a file and return list of lines"""
    if os.path.exists(file):
        with open(file, 'r') as f:
            lines = f.read().splitlines()
            f.close()
            return(lines)
    else:
        sys.exit('%s not found' % file)


def tsv_to_json(inFile, outName, outPath):
    """ Convert .tsv output to .json output
    """
    json_dict = {}
    tsv_file = read_file(inFile)
    for line in tsv_file:
        if not line.startswith('Seed'):
            tmp = line.split()
            json_dict[f'{tmp[0]}_{tmp[1]}'] = [f'{tmp[2].split("/")[0]}', f'{tmp[2].split("/")[1]}']

    final_out = f'{os.path.abspath(outPath)}/{outName}.json'
    if not os.path.exists(final_out):
        with open(final_out, "w") as f:
            json.dump(json_dict, f)
    else:
        sys.exit(f'Output file {final_out} exists!')


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument("-i", "--input", default='.', type=str, required=True,
                          help="Input file in TSV format")
    optional.add_argument("-n", "--outName", default=None, type=str, required=False,
                          help="Name of the output json")
    optional.add_argument("-o", "--outPath", default='.', type=str, required=False,
                          help="Path to output directory")
    args = parser.parse_args()

    if not args.outName:
        outName = args.input.split('/')[-1]
    else:
        outName = args.outName
    tsv_to_json(args.input, outName, args.outPath)
    print(f'DONE! Final output saved in {os.path.abspath(args.outPath)}/{outName}.json')


if __name__ == '__main__':
    main()
